var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import path from 'path';
import { pathToFileURL, fileURLToPath } from 'url';
import 'resolve-global';
import { moduleResolve } from 'import-meta-resolve';
import mergeWith from 'lodash.mergewith';
import { validateConfig } from '@commitlint/config-validator';
import importFresh from 'import-fresh';
const dynamicImport = async (id) => {
    const imported = await import(path.isAbsolute(id) ? pathToFileURL(id).toString() : id);
    return ('default' in imported && imported.default) || imported;
};
/**
 * @see moduleResolve
 */
export const resolveFrom = (specifier, parent) => {
    let resolved;
    let resolveError;
    for (const suffix of ['', '.js', '.json', '/index.js', '/index.json']) {
        try {
            resolved = moduleResolve(specifier + suffix, pathToFileURL(parent ? parent : import.meta.url));
            return fileURLToPath(resolved);
        }
        catch (err) {
            if (!resolveError) {
                resolveError = err;
            }
        }
    }
    throw resolveError;
};
/**
 *
 * @param resolvedParserPreset path resolved by {@link resolveFrom}
 * @returns path and parserOpts function retrieved from `resolvedParserPreset`
 */
export const loadParserPreset = async (resolvedParserPreset) => {
    const finalParserOpts = await dynamicImport(resolvedParserPreset);
    const relativeParserPath = path.relative(process.cwd(), resolvedParserPreset);
    return {
        path: `./${relativeParserPath}`.split(path.sep).join('/'),
        parserOpts: finalParserOpts,
    };
};
export default async function resolveExtends(config = {}, context = {}) {
    const { extends: e } = config;
    const extended = await loadExtends(config, context);
    extended.push(config);
    return extended.reduce((r, _a) => {
        var { extends: _ } = _a, c = __rest(_a, ["extends"]);
        return mergeWith(r, c, (objValue, srcValue, key) => {
            if (key === 'plugins') {
                if (Array.isArray(objValue)) {
                    return objValue.concat(srcValue);
                }
            }
            else if (Array.isArray(objValue)) {
                return srcValue;
            }
        });
    }, e ? { extends: e } : {});
}
/**
 * Fake file name to provide {@link moduleResolve} a filename to resolve from the configuration cwd
 */
const FAKE_FILE_NAME_FOR_RESOLVER = '__';
async function loadExtends(config = {}, context = {}) {
    const { extends: e } = config;
    const ext = e ? (Array.isArray(e) ? e : [e]) : [];
    return await ext.reduce(async (configs, raw) => {
        const resolved = resolveConfig(raw, context);
        const c = await (context.dynamicImport || dynamicImport)(resolved);
        const cwd = path.dirname(resolved);
        const ctx = Object.assign(Object.assign({}, context), { cwd });
        // Resolve parser preset if none was present before
        if (!context.parserPreset &&
            typeof c === 'object' &&
            typeof c.parserPreset === 'string') {
            const resolvedParserPreset = resolveFrom(c.parserPreset, path.join(cwd, FAKE_FILE_NAME_FOR_RESOLVER));
            const parserPreset = Object.assign({ name: c.parserPreset }, (await loadParserPreset(resolvedParserPreset)));
            ctx.parserPreset = parserPreset;
            config.parserPreset = parserPreset;
        }
        validateConfig(resolved, config);
        return [...(await configs), ...(await loadExtends(c, ctx)), c];
    }, Promise.resolve([]));
}
function getId(raw = '', prefix = '') {
    const first = raw.charAt(0);
    const scoped = first === '@';
    const relative = first === '.';
    const absolute = path.isAbsolute(raw);
    if (scoped) {
        return raw.includes('/') ? raw : [raw, prefix].filter(String).join('/');
    }
    return relative || absolute ? raw : [prefix, raw].filter(String).join('-');
}
function resolveConfig(raw, context = {}) {
    const resolve = context.resolve || tryResolveId;
    const id = getId(raw, context.prefix);
    let resolved;
    try {
        resolved = resolve(id, context);
    }
    catch (err) {
        const legacy = getId(raw, 'conventional-changelog-lint-config');
        resolved = resolve(legacy, context);
        console.warn(`Resolving ${raw} to legacy config ${legacy}. To silence this warning raise an issue at 'npm repo ${legacy}' to rename to ${id}.`);
    }
    return resolved;
}
function tryResolveId(id, context) {
    const cwd = context.cwd || process.cwd();
    for (const suffix of ['', '.js', '.json', '/index.js', '/index.json']) {
        try {
            return fileURLToPath(moduleResolve(id + suffix, pathToFileURL(path.join(cwd, id))));
        }
        catch (_a) { }
    }
    return resolveId(id, context);
}
function resolveId(specifier, context = {}) {
    const cwd = context.cwd || process.cwd();
    const localPath = resolveFromSilent(specifier, cwd);
    if (typeof localPath === 'string') {
        return localPath;
    }
    const resolveGlobal = context.resolveGlobal || resolveGlobalSilent;
    const globalPath = resolveGlobal(specifier);
    if (typeof globalPath === 'string') {
        return globalPath;
    }
    const err = new Error(`Cannot find module "${specifier}" from "${cwd}"`);
    throw Object.assign(err, { code: 'MODULE_NOT_FOUND' });
}
function resolveFromSilent(specifier, parent) {
    try {
        return resolveFrom(specifier, path.join(parent, FAKE_FILE_NAME_FOR_RESOLVER));
    }
    catch (err) { }
}
function resolveGlobalSilent(specifier) {
    try {
        const resolveGlobal = importFresh('resolve-global');
        return resolveGlobal(specifier);
    }
    catch (err) { }
}
//# sourceMappingURL=index.js.map